//this example is to show how to use hmbdc's request-reply (sync messaging) to write simple 
//blind auction program, a list of items are auction one by one by the Auctioner
//Bidders are running on a different process (or hosts) (bids within 1 seconds are 
//collected and best bid wins)
//
//
//to build: (newer compilers might require -faligned-new compile flag)
//g++ blind-auction.cpp -g -O3 -std=c++1y -Wall -Werror -pthread -I path-to-boost  -I /opt/hmbdc/include/ /opt/hmbdc/lib/libhmbdc.a /usr/local/lib/libboost_system.a /usr/local/lib/libboost_iostreams.a -lpthread -lrt -o /tmp/server-cluster
//
//$ /tmp/blind-auction # start the bidders first
//$ /tmp/blind-auction auction-start # start request bids for each items

#include <hmbdc/app/tcpcast/NetContext.hpp>
#include <hmbdc/os/Signals.hpp>
#include <hmbdc/time/Time.hpp>
#include <hmbdc/app/Logger.hpp>

#include <iostream>
#include <string>
#include <memory>
#include <cstdlib> 

using namespace std;
using namespace hmbdc;
using namespace hmbdc::app;
using namespace hmbdc::app::tcpcast;

char const AUCTION_TOPIC[] = "blind-auction-example";
struct AuctionItem
: hasTag<1001> {
    char description[120];
};

struct Bid
: hasTag<1004> {
    uint8_t price;
    char bidder[16];
};

using NetCtx = tcpcast::NetContext;
using AuctionContext = Context<sizeof(REQUEST<AuctionItem>)>;

//! [nework request reply example] 
template <typename Ctx>
struct Auctioner
: Client<Auctioner<Ctx>, REPLY<Bid>> {
    Auctioner(Ctx& ctx, char const ** items, size_t count)
    : ctx_(ctx)
    , sender_(NetCtx::instance().getSender(AUCTION_TOPIC))
    , items_(items)
    , count_(count) {
    }

    void invokedCb(uint16_t) {
        if (current_ == count_) throw 0; //all done

        REQUEST<AuctionItem> ra;
        strcpy(ra.request.description, items_[current_++]);
        //blocking for a second to collect all bids in this period
        //for this item
        sender_->requestReply(ra
            , ctx_
            , *this
            , time::Duration::seconds(1));
        if (highestBidder_.size()) {
            std::cout << ra.request.description << " sold at " << (int)highestPrice_
            << " to " << highestBidder_ << std::endl;
        } else {
            std::cout << ra.request.description << " bought-in " << std::endl;
        }
        highestPrice_ = 0;
        highestBidder_ = std::string();
    }
    
    //when replies come back - handle it here since 
    //the Client regular callbacks is blocked on requestReply()
    // - it is safe to use any data here
    //this Bid is garanteed by hmbdc to be for the current Item,
    //request-reply ensures the above
    void handleReplyCb(Bid const& bid) {
        if (bid.price > highestPrice_) {
            highestPrice_ = bid.price;
            highestBidder_ = std::string(bid.bidder);
        }
    }

private:
    Ctx& ctx_;
    Sender* sender_;
    char const** items_;
    size_t count_;
    size_t current_ = 0;
    uint8_t highestPrice_ = 0;
    std::string highestBidder_;
};

//! [nework request reply example] 

struct Bidder
: Client<Bidder, REQUEST<AuctionItem>> {
    Bidder(int id)
    : name_("bidder #")
    , sender_(NetCtx::instance().getSender(AUCTION_TOPIC)) {
        name_ += std::to_string(id);
    }

    void handleMessageCb(REQUEST<AuctionItem> const& item) {
        REPLY<Bid> reply(item);
        reply.reply.price = (uint8_t)(rand() % 256);
        strncpy(reply.reply.bidder, name_.c_str(), sizeof(reply.reply.bidder));
        sender_->sendReply(reply);
        std::cout << name_ << " bid with " << (int)reply.reply.price << std::endl;
    }

    std::string name_;
    Sender* sender_;
};

/**
 * @brief run in auctioner mode
 * 
 */
void runAuctioner() {
    AuctionContext ctx;      //context to power network transports and other things
    SingletonGuardian<NetCtx> g(ctx); //RAII for udpcast::NetContext resources
    auto& net = NetCtx::instance(); 
    net.listenTo(AUCTION_TOPIC);

    char const * items[] = {
        "new carpet 6x8 ft (never used)",
        "cigar box - hardwood",
        "Apple II - collectable PC 1980",
        "Some box - have to guess content",
    };

    Auctioner<AuctionContext> auctioner(ctx, items, hmbdc::length_of(items));
    hmbdc::os::HandleSignals::onTermIntDo(
        [&ctx] {
            ctx.stop();
        }
    );
    ctx.start(2, 3);
    sleep(2); //wait topic to be hot
    ctx.start(auctioner, 0x04ul);
    ctx.join();
}

void runBidders() {
    AuctionContext ctx;      //context to power network transports and other things
    SingletonGuardian<NetCtx> g(ctx); //RAII for udpcast::NetContext resources
    auto& net = NetCtx::instance(); 
    net.listenTo(AUCTION_TOPIC);
    auto bidderCount = 100;
    hmbdc::os::HandleSignals::onTermIntDo(
        [&ctx] {
            ctx.stop();
        }
    );
    std::list<std::unique_ptr<Bidder>> bidders;
    for (auto i = 0; i < bidderCount; ++i) {
        auto bidder = new Bidder{i};
        bidders.emplace_back(bidder);
        ctx.addToPool(*bidder);
    }
    ctx.start(4, 0x0f);
    ctx.join();
}

int main(int argc, char** argv) {
    // SingletonGuardian<SyncLogger> logGuard(std::cout);
    std::cout << "ctrl-c to exit ..." << std::endl;
    if (argc == 2) {
        runAuctioner();
    } else {
        runBidders();
    }

    return 0;
}

