#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Message.hpp"
#include "hmbdc/app/RequestReply.hpp"
#include "hmbdc/app/Logger.hpp"
#include "hmbdc/app/Context.hpp"
#include "hmbdc/time/Timers.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/pattern/BlockingBuffer.hpp"
#include "hmbdc/os/Thread.hpp"
#include "hmbdc/MetaUtils.hpp"
#include "hmbdc/Exception.hpp"

#include <type_traits>
#include <tuple>
#include <vector>
#include <thread>
#include <mutex>
#include <stdexcept>

namespace hmbdc { namespace app {
namespace blocking_context_detail {

template <bool is_timer_manager>
struct tm_runner {
    template<typename C>
    void operator()(C&) {}
};

template <>
struct tm_runner<true> {
    void operator()(time::TimerManager& tm) {
        tm.checkTimers(time::SysTime::now());
    }
};

template <typename... MessageTuples>
struct context_property_aggregator {
    using Interests = std::tuple<>;
    using Requests = std::tuple<>;
    using Replies = std::tuple<>;
};

template <typename MessageTuple, typename... MessageTuples>
struct context_property_aggregator<MessageTuple, MessageTuples ...> {
    using Interests = typename hmbdc::merge_tuple_unique<
        typename templatized_subtractor<REPLY, MessageTuple>::type
    	, typename context_property_aggregator<MessageTuples ...>::Interests
    >::type;

    using Requests = typename hmbdc::merge_tuple_unique<
        typename templatized_aggregator<REQUEST, MessageTuple>::type
        , typename context_property_aggregator<MessageTuples ...>::Requests
    >::type;

    using Replies = typename merge_tuple_unique<
        typename templatized_aggregator<REPLY, MessageTuple>::type
        , typename context_property_aggregator<MessageTuples ...>::Replies
    >::type;
};


template <typename T>
typename std::enable_if<std::is_base_of<time::TimerManager, T>::value, time::Duration>::type
waitDuration(T const& tm, time::Duration maxBlockingTime) {
    return std::min(tm.untilNextFire(), maxBlockingTime);
}

template <typename T>
typename std::enable_if<!std::is_base_of<time::TimerManager, T>::value, time::Duration>::type
waitDuration(T const&, time::Duration maxBlockingTime) {
    return maxBlockingTime;
}

template <typename CcClient>
bool runOnceImpl(bool& HMBDC_RESTRICT stopped
    , pattern::BlockingBuffer* HMBDC_RESTRICT buf
    , CcClient& HMBDC_RESTRICT c, time::Duration maxBlockingTime) {
    pattern::BlockingBuffer::iterator begin, end;
    try {
        tm_runner<std::is_base_of<time::TimerManager, CcClient>::value> tr;
        tr(c);

        const bool clientParticipateInMessaging = 
            std::decay<CcClient>::type::INTERESTS_SIZE != 0;
        if (clientParticipateInMessaging && buf) {
            uint64_t count = buf->peek(begin, end);
            auto b = begin;
            c.CcClient::handleRangeImpl(b, end, 0xffff);
            c.CcClient::invokedCb(0xffff);
            buf->wasteAfterPeek(count);
            if (!count) {
                buf->waitItem(waitDuration(c, maxBlockingTime));
            }
        } else {
            c.CcClient::invokedCb(0xffffu);
            auto d = waitDuration(c, maxBlockingTime);
            if (d) usleep(d.microseconds());
        }
    } catch (std::exception const& e) {
        if (!stopped) {
            c.stopped(e);
            return !c.dropped();
        }
    } catch (int code) {
        if (!stopped) {
            c.stopped(ExitCode(code));
            return !c.dropped();
        }
    } catch (...) {
        if (!stopped) {
            c.stopped(UnknownException());
            return !c.dropped();
        }
    }
    return true;
}


template <typename CcClient>
bool runOnceLoadSharingImpl(bool& HMBDC_RESTRICT stopped
    , pattern::BlockingBuffer* HMBDC_RESTRICT buf
    , CcClient& HMBDC_RESTRICT c, time::Duration maxBlockingTime) {
    pattern::BlockingBuffer::iterator begin, end;
    try {
        tm_runner<std::is_base_of<time::TimerManager, CcClient>::value> tr;
        tr(c);

        const bool clientParticipateInMessaging = 
            std::decay<CcClient>::type::INTERESTS_SIZE != 0;
        if (clientParticipateInMessaging && buf) {
            uint8_t msgWrapped[buf->maxItemSize()] __attribute__((aligned (8)));
            void* tmp = msgWrapped;
            if (buf->tryTake(tmp, 0, waitDuration(c, maxBlockingTime))) {
                MessageDispacher<CcClient, typename CcClient::Interests>()(
                    c, *static_cast<MessageHead*>(tmp));
            }
            c.CcClient::invokedCb(0xffffu);
        } else {
            c.CcClient::invokedCb(0xffffu);
            auto d = waitDuration(c, maxBlockingTime);
            if (d) usleep(d.microseconds());
        }
    } catch (std::exception const& e) {
        if (!stopped) {
            c.stopped(e);
            return !c.dropped();
        }
    } catch (int code) {
        if (!stopped) {
            c.stopped(ExitCode(code));
            return !c.dropped();
        }
    } catch (...) {
        if (!stopped) {
            c.stopped(UnknownException());
            return !c.dropped();
        }
    }
    return true;
}
} //blocking_context_detail

/**
 * @class BlockingContext<>
 * @brief A BlockingContext is like a media object that facilitates the communications 
 * for the Clients that it is holding. Each Client is powered by a single OS thread.
 * a Client needs to be started once and only once to a single BlockingContext 
 * before any messages sending happens - typically in the initialization stage in main(),
 * undefined behavior otherwise.
 * @details a Client running in such a BlockingContext utilizing OS's blocking mechanism
 * and takes less CPU time. The Client's responding time scales better when the number of 
 * Clients greatly exceeds the availlable CPUs in the system and the effective message rate
 * for a Client tends to be low.
 * @tparam MessageTuples std tuple capturing the Messages that the Context is supposed to
 * deliver. Messages that not listed here are silently dropped to ensure loose coupling
 * between senders and receivers
 */
template <MessageTupleC... MessageTuples>
struct BlockingContext {
private:
    using cpa = blocking_context_detail::context_property_aggregator<MessageTuples ...>;
    template <typename Message> struct can_handle;
public:
    /**
     * @class Interests
     * @brief a std tuple holding messages types it can dispatch except REPLYs
     * @details will not compile if using the Context to deliver a message not in this tuple
     * 
     */
    using Interests = typename cpa::Interests;

    /**
     * @class Requests
     * @brief a std tuple holding REQUEST messages types it can dispatch
     * @details will not compile if using the Context to deliver a REQUEST not in this tuple
     * 
     */
    using Requests = typename cpa::Requests;

    /**
     * @class Replies
     * @brief a std tuple holding REPLY messages types it can dispatch
     * @details will not compile if using the Context to deliver a REPLY not in this tuple
     * 
     */
    using Replies = typename cpa::Replies;


    struct deliverAll {
        template <typename T>
        bool operator()(T&&) {return true;}
    };

    /**
     * @brief trivial ctor
     */
    template <typename U = Requests>
    BlockingContext(typename std::enable_if<
        std::tuple_size<U>::value == 0, void>::type* = nullptr)
    : stopped_(false) {}

    template <typename U = Requests>
    BlockingContext(typename std::enable_if<
        std::tuple_size<U>::value, void>::type* = nullptr)
    : stopped_(false) {
        rrCtx_.start(1, 0);
    }


    BlockingContext(BlockingContext const&) = delete;
    BlockingContext& operator = (BlockingContext const&) = delete;

    /**
     * @brief start a client within the Context. The client is powered by a single OS thread.
     * @details it is ok if a Client is blocking, if its own buffer is not full, it
     * doesn't affect other Client's capabilities of receiving Messages
     * 
     * @tparam Client actual Client type
     * @tparam DeliverPred a condition functor deciding if a message should be delivered to a
     * Client, which provides filtering before the Message type filtering. It improves performance
     * in the way it could potentially reduce the unblocking times of thread
     * 
     * Usage example: 
     * //the following Pred verifies srcId matches for Response, and let all other Messages types
     * //deliver
     * @code
     * struct Pred {
     * Pred(uint16_t srcId)
     *     : srcId(srcId){}
     *    
     *     bool operator()(Response const& resp) {
     *         return resp.srcId == srcId;
     *     }
     *     template <typename M>
     *     bool operator()(M const&) {return true;}
     * 
     *     uint16_t srcId;
     * };
     * 
     * @endcode
     * 
     * @param c Client
     * @param capacity the maximum messages this Client can buffer
     * @param maxItemSize the max size of a message - when the Client is interetsed
     * in JustBytes, this value MUST be big enough to hold the max size message to
     * avoid truncation
     * @param cpuAffinity cpu affinity mask for this Client's thread
     * @param maxBlockingTime it is recommended to limit the duration a Client blocks due to 
     * no messages to handle, so it can respond to things like Context is stopped, or generate
     * heartbeats if applicable.
     * @param pred see DeliverPred documentation
     */
	template <typename Client, typename DeliverPred = deliverAll>
    void start(Client& c
        , size_t capacity = 1024
        , size_t maxItemSize = max_size_in_tuple<typename Client::Interests>::value
        , uint64_t cpuAffinity = 0
        , time::Duration maxBlockingTime = time::Duration::seconds(1)
        , DeliverPred&& pred = DeliverPred()) {
        if (maxItemSize < max_size_in_tuple<typename Client::Interests>::value)
            HMBDC_THROW(std::out_of_range, "maxItemSize is too small");
    	std::shared_ptr<Transport> t(new Transport(maxItemSize, capacity));
		setupConduit<MsgConduits, DeliverPred, typename Client::Interests>()(
            msgConduits_, t, std::forward<DeliverPred>(pred));
        auto thrd = kickOffClientThread(blocking_context_detail::runOnceImpl<Client>
            , c, t.use_count() == 1?nullptr:&t->buffer, cpuAffinity, maxBlockingTime);
        threads_.push_back(move(thrd));
    }

    /**
     * @brief start a group of clients within the Context that collectively processing 
     * messages in a load sharing manner. Each client is powered by a single OS thread
     * @details it is ok if a Client is blocking, if its own buffer is not full, it
     * doesn't affect other Client's capabilities of receiving Messages
     * 
     * @tparam LoadSharingClientPtrIt iterator to a Client pointer
     * @tparam DeliverPred a condition functor deciding if a message should be delivered to these
     * Clients, which provides filtering before the Message type filtering. It improves performance
     * in the way it could potentially reduce the unblocking times of threads.
     * An exception is for Client that are interested in JustBytes (basically every message),
     * this filter does not apply - no filtering for these Clients
     * 
     * @param begin an iterator, **begin should produce a Client& for the first Client
     * @param end an end iterator, [begin, end) is the range for Clients
     * @param capacity the maximum messages this Client can buffer
     * @param maxItemSize the max size of a message - when the Client is interetsed
     * in JustBytes, this value MUST be big enough to hold the max size message to
     * avoid truncation
     * @param cpuAffinity cpu affinity mask for this Client's thread
     * @param maxBlockingTime it is recommended to limit the duration a Client blocks due to 
     * no messages to handle, so it can respond to things like Context is stopped, or generate
     * heartbeats if applicable.
     * @param pred see DeliverPred documentation
     */
    template <typename LoadSharingClientPtrIt, typename DeliverPred = deliverAll>
    void start(LoadSharingClientPtrIt begin, LoadSharingClientPtrIt end
        , size_t capacity = 1024
        , size_t maxItemSize = max_size_in_tuple<
            typename std::decay<decltype(**LoadSharingClientPtrIt())>::type::Interests
        >::value
        , uint64_t cpuAffinity = 0
        , time::Duration maxBlockingTime = time::Duration::seconds(1)
        , DeliverPred&& pred = DeliverPred()) { //=[](auto&&){return true;}
        using Client = typename std::decay<
                decltype(**LoadSharingClientPtrIt())
            >::type;
        if (maxItemSize < max_size_in_tuple<typename Client::Interests>::value) 
            HMBDC_THROW(std::out_of_range, "maxItemSize is too small");
        if (begin == end) return;
        std::shared_ptr<Transport> t(new Transport(maxItemSize, capacity));
        setupConduit<MsgConduits, DeliverPred, typename Client::Interests>()(
            msgConduits_, t, std::forward<DeliverPred>(pred));
        while(begin != end) {
            auto thrd = kickOffClientThread(blocking_context_detail::runOnceLoadSharingImpl<Client>
                , **begin++, t.use_count() == 1?nullptr:&t->buffer, cpuAffinity, maxBlockingTime);
            threads_.push_back(std::move(thrd));
        }
    }

    /**
     * @brief stop the message dispatching - asynchronously
     * @details asynchronously means not garanteed message dispatching 
     * stops immidiately after this non-blocking call
     */
    void stop() {
        __atomic_thread_fence(__ATOMIC_ACQUIRE);
        stopped_ = true;
    }

    /**
     * @brief wait until all threads of the Context exit
     * @details blocking call
     */
    void join() {
        for (auto& t : threads_) {
            t.join();
        }
        threads_.clear();
    }

    /**
     * @brief send a message to the BlockingContext to dispatch
     * @details only the Clients that handles the Message will get it
     * This function is threadsafe, which means you can call it anywhere in the code
     * 
     * @tparam Message type
     * @param m message
     */
    template <MessageC Message>
    typename std::enable_if<can_handle<Message>::match, void>::type
    send(Message&& m) {
        justBytesSend(m);
        using M = typename std::decay<Message>::type;
        auto constexpr i = index_in_tuple<TransportEntries<M>, MsgConduits>::value;
        static_assert(i < std::tuple_size<MsgConduits>::value, "unexpected message type");
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (e.diliverPred(m)) {
                if (i == entries.size() - 1) {
                    e.transport->buffer
                        .template putInPlace<MessageWrap<M>>(std::forward<Message>(m));
                } else {
                    e.transport->buffer.template putInPlace<MessageWrap<M>>(m);
                }
            }
        }
    }
    template <MessageC Message>
    typename std::enable_if<!can_handle<Message>::match, void>::type
    send(Message&& m){
        justBytesSend(std::forward<Message>(m));
        if (!can_handle<Message>::justbytes) {
            HMBDC_LOG_ONCE(HMBDC_LOG_D("unhandled message sent typeTag="
                , std::decay<Message>::type::typeTag);)
        }
    }
    
    /**
     * @brief send a message by directly constructing the message in receiving message queue
     * @details since the message only exists after being deleivered, the DeliverPred
     * is not called to decide if it should be delivered
     * 
     * @param args ctor args
     * @tparam Message type
     * @tparam typename ... Args args
     */
    template <MessageC Message, typename ... Args>
    typename std::enable_if<can_handle<Message>::match, void>::type
    sendInPlace(Args&&... args) {
        justBytesSendInPlace<Message>(args...);
        using M = typename std::decay<Message>::type;
        auto constexpr i = index_in_tuple<TransportEntries<M>, MsgConduits>::value;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (i == entries.size() - 1) {
                e.transport->buffer
                    .template putInPlace<MessageWrap<M>>(std::forward<Args>(args)...);
            } else {
                e.transport->buffer.template putInPlace<MessageWrap<M>>(args...);
            }
        }
    }
    template <MessageC Message, typename ... Args>
    typename std::enable_if<!can_handle<Message>::match, void>::type
    sendInPlace(Args&& ...args) {
        justBytesSendInPlace<Message>(std::forward<Args>(args)...);
        if (!can_handle<Message>::justbytes) {
            HMBDC_LOG_ONCE(HMBDC_LOG_D("unhandled message sent typeTag="
                , std::decay<Message>::type::typeTag);)
        }
    }

    /**
     * @brief best effort to send a message to via the BlockingContext
     * @details this method is not recommended if more than one recipients are accepting
     * this message since there is no garantee each one will receive it once and only once.
     * this call does not block - return false when deliver doesn't reach all
     * intended recipients
     * This method is threadsafe, which means you can call it anywhere in the code
     * 
     * @param m message
     * @param timeout return false if cannot deliver in the specified time
     * @return true if send successfully to every intended receiver
     */
    template <MessageC Message>
    typename std::enable_if<can_handle<Message>::match, bool>::type
    trySend(Message&& m, time::Duration timeout = time::Duration::seconds(0)) {
        if (!justBytesTrySend(m)) return false;
        using M = typename std::decay<Message>::type;
        auto constexpr i = index_in_tuple<TransportEntries<M>, MsgConduits>::value;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (e.diliverPred(m)) {
                if (i == entries.size() - 1) {
                    if (!e.transport->buffer
                        .tryPut(MessageWrap<M>(std::forward<Message>(m)), timeout)) {
                        return false;
                    }
                } else {
                    if (!e.transport->buffer.tryPut(MessageWrap<M>(m), timeout)) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    template <MessageC Message>
    typename std::enable_if<!can_handle<Message>::match, bool>::type
    trySend(Message&& m, time::Duration = time::Duration::seconds(0)) {
        if (!can_handle<Message>::justbytes) {
            HMBDC_LOG_ONCE(HMBDC_LOG_D("unhandled message sent typeTag="
                , std::decay<Message>::type::typeTag);)
        }

        return justBytesTrySend(std::forward<Message>(m));
    }

   
    /**
     * @brief best effort to send a message by directly constructing the message in 
     * receiving message queue
     * @details since the message only exists after being deleivered, the DeliverPred
     * is not called to decide if it should be delivered
     * this method is not recommended if more than one recipients are accepting
     * this message since it is hard to ensure each message is delivered once and only once.
     * this call does not block - return false when delivery doesn't reach ALL
     * intended recipients
     * This method is threadsafe, which means you can call it anywhere in the code
     * 
     * @return true if send successfully to every intended receiver
     */
    template <MessageC Message, typename ... Args>
    typename std::enable_if<can_handle<Message>::match, bool>::type
    trySendInPlace(Args&&... args) {
        if (!justBytesTrySendInPlace<Message>(args...)) return false;
        using M = typename std::decay<Message>::type;
        auto constexpr i = index_in_tuple<TransportEntries<M>, MsgConduits>::value;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (i == entries.size() - 1) {
                if (!e.transport->buffer.template 
                    tryPutInPlace<MessageWrap<M>>(std::forward<Args>(args)...)) {
                    return false;
                }
            } else {
                if (!e.transport->buffer.template 
                    tryPutInPlace<MessageWrap<M>>(args...)) {
                    return false;
                }
            }
        }
        return true;
    }

    template <MessageC Message, typename ... Args>
    typename std::enable_if<!can_handle<Message>::match, bool>::type
    trySendInPlace(Args&&... args) {
        if (!can_handle<Message>::justbytes) {
            HMBDC_LOG_ONCE(HMBDC_LOG_D("unhandled message sent typeTag="
                , std::decay<Message>::type::typeTag);)
        }

        return justBytesTrySendInPlace<Message>(args...);
    }

   /**
     * @brief send a range of messages via the BlockingContext 
     * - only checking with diliverPred using the first message
     * @details only the Clients that handles the Message will get it of course
     * This function is threadsafe, which means you can call it anywhere in the code
     * 
     * @param begin a forward iterator point at the start of the range 
     * @param n length of the range
     */
    template <MessageForwardIterC ForwardIt>
    typename std::enable_if<can_handle<decltype(*(ForwardIt()))>::match, void>::type
    send(ForwardIt begin, size_t n) {
        if (!n) return;
        justBytesSend(begin, n);
        using Message = decltype(*begin);
        using M = typename std::decay<Message>::type;
        auto constexpr i = index_in_tuple<TransportEntries<M>, MsgConduits>::value;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (e.diliverPred(*begin)) {
                while (!e.transport->buffer.template 
                    tryPutBatchInPlace<MessageWrap<M>>(begin, n)) {}
            }
        }
    }

    template <MessageForwardIterC ForwardIt>
    typename std::enable_if<!can_handle<decltype(*(ForwardIt()))>::match, void>::type
    send(ForwardIt begin, size_t n) {
        if (!can_handle<decltype(*(ForwardIt()))>::justbytes) {
            HMBDC_LOG_ONCE(HMBDC_LOG_D("unhandled message sent typeTag="
                , std::decay<decltype(*(ForwardIt()))>::type::typeTag);)
        }
        justBytesSend(begin, n);
    }

    /**
     * @brief part of synchronous request reply interface. The caller is blocked until
     * the replies to this request are heard and processed. Just liek regular Messages
     * REQUEST is sent out to Clients that are interested in this Context 
     * 
     * @details 
     * 
     * @tparam MessageC Message the type wrapped in REQUEST<> template
     * @tparam typename ReplyHandler type for how to handle the replies.
     *      example of a handler expect two steps replies for requestReply call:
     *      @code
     *      struct ClientAlsoHandlesReplies
     *      : Client<ClientAlsoHandlesReplies, REPLY<Reply0>, REPLY<Reply1>, OtherMsg> {
     *          void handleReplyCb(Reply0 const&){
     *              // got step 1 reply
     *          }
     *          void handleReplyCb(Reply1 const&){
     *              // got step 2 reply 
     *              throw hmbdc::ExitCode(0); // all replies received, now
     *                                        // throw to unblock the requestReply call
     *          }
     *          @endcode
     * 
     * @param request the REQUEST sent out
     * @param replyHandler the callback logic to handle replies
     * @param timeout timeout value
     * @return true if replies are received and handled, false if timedout
     */
    template <MessageC M, typename ReplyHandler>
    bool
    requestReply(REQUEST<M>&& request, ReplyHandler& replyHandler
        , time::Duration timeout = time::Duration::seconds(0xffffffff)) {
        static_assert(std::tuple_size<Requests>::value != 0
            , "no REQUEST is declared for this Context");
        return hmbdc::app::requestReply(std::forward<REQUEST<M>>(request)
            , *this
            , rrCtx_
            , replyHandler
            , timeout);
    }

    /**
     * @brief part of synchronous request reply interface. Send a constructed REPLY
     * to the Context, it will only reach to the origianl sender Client. 
     * ALL REPLY need to sent out using this interface to avoid
     * undefined behavior.
     * 
     * @tparam MessageC Message the type wrapped in REPLY<> template
     * @param reply the rvalue REPLY (previously constructed from a REQUEST) to sent out
     */
    template <ReplyC ReplyT>
    void
    sendReply(ReplyT &&reply) {
        using R = typename std::decay<ReplyT>::type;
        static_assert(std::is_trivially_destructible<R>::value, "unsupported");
        static_assert(std::tuple_size<Requests>::value != 0
            , "no REQUEST is declared for this Context, no one would be waiting");
        rrCtx_.send(std::forward<ReplyT>(reply));
    }

private:
    void start(){}

    struct Transport {
    	Transport(size_t maxItemSize, size_t capacity)
    	: buffer(maxItemSize + sizeof(MessageHead), capacity){}
    	pattern::BlockingBuffer buffer;
    };
     
    template <typename Message>
    struct TransportEntry {
    	std::shared_ptr<Transport> transport;
        std::function<bool (Message const&)> diliverPred;
    };

    template <typename Message>
    using TransportEntries = std::vector<TransportEntry<Message>>;

    template <typename Tuple> struct MCGen;
    template <typename ...Messages>
    struct MCGen<std::tuple<Messages ...>> {
        using type = std::tuple<TransportEntries<Messages> ...>;
    };

    using MsgConduits = typename MCGen<Interests>::type;

    MsgConduits msgConduits_;
    using Threads = std::vector<std::thread>;
    Threads threads_;
    bool stopped_;
    typename std::conditional<std::tuple_size<Requests>::value
        , Context<max_size_in_tuple<Replies>::value
            , context_property::broadcast<4>
        >
        , void*
    >::type rrCtx_;

    template <typename MsgConduits, typename DeliverPred, typename Tuple>
    struct setupConduit {
        void operator()(MsgConduits&, std::shared_ptr<Transport>, DeliverPred&&){}
    };

    struct createEntry {
        template <bool careMsg, size_t i, typename M, typename MsgConduits, typename DeliverPred>
        typename std::enable_if<careMsg, void>::type
        doOrNot(MsgConduits& msgConduits, std::shared_ptr<Transport> t, DeliverPred&& pred) {
            auto& entries = std::get<i>(msgConduits);
            entries.emplace_back(TransportEntry<M>{t, pred});
        }

        template <bool careMsg, size_t i, typename M, typename MsgConduits, typename DeliverPred>
        typename std::enable_if<!careMsg, void>::type
        doOrNot(MsgConduits&, std::shared_ptr<Transport>, DeliverPred&&) {
        }
    };

    template <typename MsgConduits, typename DeliverPred, typename M, typename ...Messages>
    struct setupConduit<MsgConduits, DeliverPred, std::tuple<M, Messages...>> {
        void operator()(MsgConduits& msgConduits, std::shared_ptr<Transport> t, DeliverPred&& pred){
            auto constexpr i = index_in_tuple<M, Interests>::value;
            createEntry().template doOrNot<i != std::tuple_size<Interests>::value, i, M>(
                msgConduits, t, pred);
        	setupConduit<MsgConduits, DeliverPred, std::tuple<Messages...>>()
                (msgConduits,t, std::forward<DeliverPred>(pred));
        }
    };

    template <typename MsgConduits, typename M, typename ...Messages>
    struct setupConduit<MsgConduits, void*, std::tuple<M, Messages...>> {
        void operator()(MsgConduits& msgConduits, std::shared_ptr<Transport> t, void*){
            auto constexpr i = index_in_tuple<M, Interests>::value;
            createEntry().template doOrNot<i != std::tuple_size<Interests>::value, i, M>(
                msgConduits, t, [](M const&){return true;});
            setupConduit<MsgConduits, void*, std::tuple<Messages...>>()
                (msgConduits, t, nullptr);
        }
    };

    template <typename Message>
    struct can_handle {
        private:
            using M = typename std::decay<Message>::type;
        public:
            enum {
                index = index_in_tuple<TransportEntries<M>, MsgConduits>::value,
                match = index < std::tuple_size<MsgConduits>::value,
                justbytes_index = index_in_tuple<TransportEntries<JustBytes>, MsgConduits>::value,
                justbytes = justbytes_index < std::tuple_size<MsgConduits>::value 
                    && std::is_trivially_destructible<M>::value? 1 :0,
            };
    };

	template <typename Client, typename RunOnceFunc>
    auto kickOffClientThread(RunOnceFunc runOnceFunc, Client& c, pattern::BlockingBuffer* buffer
        , uint64_t mask, time::Duration maxBlockingTime) {
        if (buffer == nullptr) {
            HMBDC_LOG_ONCE(HMBDC_LOG_D("starting a Client that cannot receive messages, Client typid()="
                , typeid(Client).name());)
        }
        std::thread thrd([
            this
            , &c
            , buffer
            , mask
            , maxBlockingTime
            , runOnceFunc]() {
            std::string name;
            char const* schedule;
            int priority;
            
            if (c.hmbdcName()) {
                name = c.hmbdcName();
            } else {
                name = "hmbdc-b";
            }
            auto cpuAffinityMask = mask; 
            std::tie(schedule, priority) = c.schedSpec();

            if (!schedule) schedule = "SCHED_OTHER";

            os::configureCurrentThread(name.c_str(), cpuAffinityMask
                , schedule, priority);
            c.messageDispatchingStartedCb(0xffff);

            while(!stopped_ && runOnceFunc(this->stopped_, buffer, c, maxBlockingTime)) {
            }
            if (this->stopped_) c.dropped();
        }
        );

        return thrd;
    }


    template <MessageC Message>
    typename std::enable_if<can_handle<Message>::justbytes, void>::type
    justBytesSend(Message&& m) {
        using M = typename std::decay<Message>::type;
        auto constexpr i = can_handle<Message>::justbytes_index;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (e.transport->buffer.maxItemSize() >= sizeof(MessageWrap<M>)) {
                e.transport->buffer.template putInPlace<MessageWrap<M>>(m);
            } else {
                HMBDC_THROW(std::out_of_range
                    , "need to increase maxItemSize in start() to >= "
                    << sizeof(MessageWrap<M>));
            }
        }
    }

    template <MessageC Message>
    typename std::enable_if<!can_handle<Message>::justbytes, void>::type
    justBytesSend(Message&& m) {
    }

    template <MessageC Message, typename ... Args>
    typename std::enable_if<can_handle<Message>::justbytes, void>::type
    justBytesSendInPlace(Args&& ...args) {
        using M = typename std::decay<Message>::type;
        auto constexpr i = can_handle<Message>::justbytes_index;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (e.transport->buffer.maxItemSize() >= sizeof(MessageWrap<M>)) {
                e.transport->buffer.template putInPlace<MessageWrap<M>>(args...);
            } else {
                HMBDC_THROW(std::out_of_range
                    , "need to increase maxItemSize in start() to >= "
                    << sizeof(MessageWrap<M>));
            }
        }
    }

    template <MessageC Message, typename ... Args>
    typename std::enable_if<!can_handle<Message>::justbytes, void>::type
    justBytesSendInPlace(Args&&... args) {
    }

    template <MessageC Message>
    typename std::enable_if<can_handle<Message>::justbytes, bool>::type
    justBytesTrySend(Message&& m, time::Duration timeout) {
        using M = typename std::decay<Message>::type;
        auto constexpr i = can_handle<Message>::justbytes_index;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (e.transport->buffer.maxItemSize() >= sizeof(MessageWrap<M>)) {
                if (!e.transport->buffer.tryPut(MessageWrap<M>(m), timeout)) {
                    return false;
                }
            } else {
                HMBDC_THROW(std::out_of_range
                    , "need to increase maxItemSize in start() to >= "
                    << sizeof(MessageWrap<M>));
            }
        }
        return true;
    }
    template <MessageC Message>
    typename std::enable_if<!can_handle<Message>::justbytes, bool>::type
    justBytesTrySend(Message&& m, time::Duration timeout) {
        return true;
    }

    template <MessageC Message, typename ... Args>
    typename std::enable_if<can_handle<Message>::justbytes, bool>::type
    justBytesTrySendInPlace(Args&&... args) {
        using M = typename std::decay<Message>::type;
        auto constexpr i = can_handle<Message>::justbytes_index;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];
            if (e.transport->buffer.maxItemSize() >= sizeof(MessageWrap<M>)) {
                if (!e.transport->buffer.template 
                    tryPutInPlace<MessageWrap<M>>(args...)) {
                    return false;
                }
            } else {
                HMBDC_THROW(std::out_of_range
                    , "need to increase maxItemSize in start() to >= "
                    << sizeof(MessageWrap<M>));
            }
        }
        return true;
    }
    template <MessageC Message, typename ... Args>
    typename std::enable_if<!can_handle<Message>::justbytes, bool>::type
    justBytesTrySendInPlace(Args&&... args) {
        return true;
    }

    template <MessageForwardIterC ForwardIt>
    typename std::enable_if<can_handle<decltype(*(ForwardIt()))>::justbytes, void>::type
    justBytesSend(ForwardIt begin, size_t n) {
        using Message = decltype(*begin);
        using M = typename std::decay<Message>::type;
        auto constexpr i = can_handle<decltype(*(ForwardIt()))>::justbytes_index;
        auto& entries = std::get<i>(msgConduits_);
        for (auto i = 0u; i < entries.size(); ++i) {
            auto& e = entries[i];

            if (e.transport->buffer.maxItemSize() >= sizeof(MessageWrap<M>)) {
                while (!e.transport->buffer.template tryPutBatchInPlace<MessageWrap<M>>(begin, n)) {}
            } else {
                HMBDC_THROW(std::out_of_range
                    , "need to increase maxItemSize in start() to >= "
                    << sizeof(MessageWrap<M>));
            }
        }
    }

    template <MessageForwardIterC ForwardIt>
    typename std::enable_if<!can_handle<decltype(*(ForwardIt()))>::justbytes, void>::type
    justBytesSend(ForwardIt begin, size_t n) {
    }

};
}}
