#include "hmbdc/Copyright.hpp"
#pragma once
 
#include "hmbdc/app/MessageDispacher.hpp"
#include "hmbdc/MetaUtils.hpp"
#include "hmbdc/Compile.hpp"

#include <iostream>
#include <tuple>
#include <type_traits>

namespace hmbdc { namespace app {

/**
 * @class single_thread_powered_client
 * @brief a trait class, if a Client can only run on a single specific thread in Pool, 
 * derive the Client from it, hmbdc will check to ensure that is the case
 */
struct single_thread_powered_client{};

/**
 * @class Client<>
 * @brief A Client represents a thread of execution/a task. The execution is managed by 
 * a Context.
 * a Client object could participate in message dispatching as the receiver of specifed 
 * message types
 * @details message is dispatched thru callbacks in the loose 
 * form of void handleMessageCb(Message const& m) or void handleMessageCb(Message& m).
 * some callbacks have default implementations.
 * however, all callbacks are overridable to provide desired effects;
 * When running in a Context, all callbacks (*Cb methods) for a hmbdc Client instance 
 * are garanteed to be called from a single OS thread, ie. no 2 callabacks of an Client 
 * instance are called concurrently, so the Client programmer can assume a 
 * single thread programming model callback-wise. 
 * The above also holds true for timer firing callbacks when a concrete Client deriving 
 * from TimerManager that manages the timers
 * 
 * See in example hmbdc.cpp @snippet hmbdc.cpp write a Client
 * 
 * @tparam CcClient the concrete Client type
 * @tparam typename ... Messages message types that the Client interested
 * if the type of JustBytes is declared, use a special callback see below
 * 
 * @code
 * void handleJustBytesCb(uint16_t tag, uint8_t* bytes){...}
 * @endcode
 * is expected
 * see ConsoleClient.hpp
 * 
 * There are 3 categories of Messages can be specified here regular, REQUEST, REPLY
 * REQUEST, REPLY are for synchronous request reply messages, this is like 
 * remote procedure call (RPC) function except it is more powerful that it gives 
 * more flexible forms of RPC returning behavior - see requestReply() sendReply() 
 * in Context.hpp, and RequestReply.hpp for details
 */
template <typename CcClient, MessageC ... Messages>
struct Client {
    using Interests
        = typename templatized_subtractor<REPLY, std::tuple<Messages ...>>::type;
    using Requests 
        = typename templatized_aggregator<REQUEST, std::tuple<Messages ...>>::type;
    using Replies 
        = typename templatized_aggregator<REPLY, std::tuple<Messages ...>>::type;

    enum {
        INTERESTS_SIZE = std::tuple_size<Interests>::value,
    };
    // static_assert(std::is_base_of<Client, CcClient>::value, "");

    /**
     * @brief return the name of thread that runs this client, override if necessary
     * @details this only used when the Client is running in direct mode
     * @return thread name - default to be hmbdc0, hmbdc1 ...
     */
    char const* hmbdcName() const { return "hmbdc-anonymous"; }

    /**
     * @brief an overrideable method.
     * returns the schedule policy and priority, override if necessary
     * priority is only used when policy is "SCHED_RR", or "SCHED_FIFO"
     * @details this is only used when the Client is running in direct mode
     * supported policy are "SCHED_OTHER"(=nullptr), "SCHED_RR", "SCHED_FIFO"
     * @return a tuple made of schedule policy and priority, default to be SCHED_OTHER
     */
    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple<char const*, int>(nullptr, 0);
    }

   /**
    * @brief an overridable method.
    * client  receives events in batches and the max batch size is controllable when running
    * in direct mode Context. Here is to specify the max size.
    * @details a message could only be reaching one client when using partition Context. In this
    * case, reduce the size (reduce the greediness) might be useful
    * @return the max number of messages when grabbing messages to process
    */
    size_t maxBatchMessageCount() const {return std::numeric_limits<size_t>::max();}


    /**
     * @brief called before any messages got dispatched - only once
     * @details this is the place some preparation code goes to
     * 
     * @param threadSerialNumber normally the number
     * indicating which thread is in action, except when INTERESTS_SIZE == 0 
     * it is another undefined value
     * 
     */
    virtual void messageDispatchingStartedCb(uint16_t threadSerialNumber) {
        //default do nothing
    };

    /**
     * @brief callback called when this Client is taken out of message dispatching
     * @details after this call the Client is still at hook from the Context point of view
     * (until droppedCb is called), so don't delete this Client yet or add it back to 
     * the Context. any exception thrown here is ignored, 
     * 
     * @param e the exception that caused the Client to be taken out of message dispatching
     * e could be thrown by the Client itself in a callback function 
     * to voluntarily take itself out
     */
    virtual void stoppedCb(std::exception const& e) {
        // called when this client is stopped, e.what() might have reason
        std::cerr << "Client " << hmbdcName() 
            << " exited with reason: " << e.what() << std::endl;
    };

    /**
     * @brief callback called after the Client is safely taken out of the Context
     * @details exception thrown here is ignored and return true is assumed
     * @return if false, this Client is added back to the Context to process messages
     * otherwise, no more callback. You could even safely "delete this; return true;"
     */
    virtual bool droppedCb() {
        // called when this client is dropped and free to be deleted
        return true;
    };
    
    /**
     * @brief this callback is called all the time (frequently) - the exact timing is
     * after a batch of messages are dispatched. After this call returns, the previously 
     * dispatched message's addresses are no longer valid, which means if you cache the 
     * event addresses in the previous handleMessageCb()s, you cannot use those after 
     * the return of the next invokeCb function.
     * @details you can collectively process the messages received/cached so far here, or 
     * do something needs to be done all the time like powering another message loop
     * 
     * @param threadSerialNumber the number 
     * indicating which thread is in action 
     */
    virtual void invokedCb(uint16_t threadSerialNumber) {
        // called as frequently as workload allowed
    }
    
    /**
     * @brief trivial
     */
    virtual ~Client(){}

protected:
    /**
     * @brief the derived user's Client has the option to stop the current batch of event 
     * dispatching.
     * @details for example, if the user's Client decides it has handled all interested 
     * messages and wants to skip the remaining messages within the CURRENT batch, call
     * this within any callback functions.
     * use with caution, only when it is absolutely certain that there is NO messages of 
     * ANY type the Client need to care are within the current batch.
     */
    void batchDone() {
        batchDone_ = true;
    }

public:
    /**
     * @brief the following are for internal use, don't change or override
     */
    void stopped(std::exception const&e) noexcept {
        try {
            stoppedCb(e);
        } catch (...) {}
    }

    bool dropped() noexcept {
        bool res = true;
        try {
            res = droppedCb();
        } catch (...) {}
        return res;
    }

    template <typename Iterator>
    void handleRangeImpl(Iterator it, 
        Iterator end, uint16_t threadId) {
        CcClient& c = static_cast<CcClient&>(*this);
        for (;hmbdc_likely(!batchDone_ && it != end); ++it) {
            MessageDispacher<CcClient, Interests>()(
                c, *static_cast<MessageHead*>(*it));
        }
        batchDone_ = false;
    }

protected:
    bool batchDone_ = false;  ///not part of API, do not touch
};

}} //hmbdc::app
