#include "hmbdc/Copyright.hpp"
#pragma once

#include "hmbdc/comm/Uuid.hpp"
#include "hmbdc/time/Time.hpp"
#include "hmbdc/Config.hpp"
#include "hmbdc/Endian.hpp"
#include "hmbdc/MetaUtils.hpp"

#include <stdint.h>
#include <utility>
#include <ostream>
#include <type_traits>
#include <bitset>
#include <memory.h>

struct epoll_event;
namespace hmbdc { namespace app {

#if __GNUC_PREREQ(9,0)

template <typename M>
concept MessageC = std::decay<M>::type::typeTag >= 0x0
    && std::decay<M>::type::typeTag <= 0xffffu;

namespace message_detail {
template <typename ...M>
constexpr bool is_message_tuple = false;

template <MessageC ...M>
constexpr bool is_message_tuple<std::tuple<M...>> = true;
}

template <typename MT>
concept MessageTupleC = message_detail::is_message_tuple<MT>;

template <typename MIter>
concept MessageForwardIterC = MessageC<decltype(*(MIter{}))> 
    && requires(MIter it) {++it;};
    
#else

#define MessageC typename
#define MessageTupleC typename
#define MessageForwardIterC typename

#endif

/** 
* see @example hmbdc.cpp
* @example perf-tcpcast.cpp
*/

/**
 * @class hasTag<>
 * @brief each message type has 16 bit tag
 * @tparam tag 16 bit unsigned tag
 */
template <uint16_t tag>
struct hasTag {
    enum {typeTag = tag};
    friend 
    std::ostream& 
    operator << (std::ostream & os, hasTag const& t) {
        return os << "hasTag=" << tag << std::endl;
    }
};
 
/** 
 * @class MessageHead
 * @details It is recommneded that a Message type do not have significant dtor defined
 * because the message could travel through process bundaries or through network 
 * - won't compile if trying to send them.
 * The only exception is the BlockingContext (licensed package), which does allow  
 * message's dtors and call them properly within a process boundary.
 * @snippet hmbdc.cpp define a message
 */
struct MessageHead {
	MessageHead(uint16_t typeTag)
	: reserved2(0)
    , reserved(0)
	, typeTag(typeTag)
	{}
	template <typename Message> Message& get();
	template <typename Message> Message const& get() const;
	template <typename Message> void setPayload(Message const&);
	template <typename Message, typename ...Args> void setPayloadInPlace(Args&& ... args);
    uint32_t reserved2;
    uint16_t reserved;

    XmitEndian<uint16_t> typeTag;
   
    using Scratchpad = 
        union {
            XmitEndian<uint16_t> payloadLen;
            XmitEndianByteField<uint64_t, 6> seq;
            uint64_t reserved : 48;
            uint8_t space[6];
        } __attribute__((packed));
    static_assert(sizeof(Scratchpad) == sizeof(reserved2) + sizeof(reserved), "");
    static_assert(sizeof(Scratchpad::seq) == sizeof(reserved2) + sizeof(reserved), "");
    Scratchpad& scratchpad() {return *reinterpret_cast<Scratchpad*>(&reserved2);}
    Scratchpad const& scratchpad() const{return *reinterpret_cast<Scratchpad const*>(&reserved2);}
    friend 
    std::ostream& operator << (std::ostream& os, MessageHead const & h) {
    	return os << h.scratchpad().reserved << ' ' << h.typeTag;
    }
} __attribute__((packed));
static_assert(sizeof(MessageHead) == 8, "pakcing issue?");

template <typename Message>
struct MessageWrap : MessageHead {
    template <typename ...Args>
	MessageWrap(Args&& ... args)
	: MessageHead(Message::typeTag)
	, payload(std::forward<Args>(args)...) {
        static_assert(std::is_base_of<hasTag<Message::typeTag>, Message>::value
            , "not a properly defined Message");
	}

    MessageWrap(MessageWrap const&) = default;
    MessageWrap(MessageWrap &) = default;
    MessageWrap(MessageWrap &&) = default;
    MessageWrap& operator = (MessageWrap const&) = default;
    MessageWrap& operator = (MessageWrap &&) = default;

    Message payload;

    friend 
    std::ostream& operator << (std::ostream& os, MessageWrap<Message> const & w) {
    	return os << static_cast<MessageHead const&>(w) << ' ' << w.payload;
    }
};

template <typename Message>
Message& 
MessageHead::get() {
	return static_cast<MessageWrap<Message>*>(this)->payload;
}

template <typename Message>
Message const& 
MessageHead::get() const{
	return static_cast<MessageWrap<Message> const*>(this)->payload;
}

template <typename Message> 
void 
MessageHead::
setPayload(Message const& m) {
	new (&get<Message>()) Message(m);
	typeTag = Message::typeTag;
}

template <typename Message, typename ...Args> 
void
MessageHead::
setPayloadInPlace(Args&& ... args) {
	new (&get<Message>()) Message(std::forward<Args>(args)...);
	typeTag = Message::typeTag;
}

struct Flush
: hasTag<0> {
};

template <size_t MaxStreamableSize>
struct LoggingT
: hasTag<3> {
    char payload[MaxStreamableSize];
} __attribute__((__may_alias__, packed));

/**
 * @class JustBytes
 * @brief A special type of message only used on the receiving side
 * @details When a Client declare this type among its interested message types, all
 * messages types (WITHOUT the destructor) will be delievered in 
 * byte form to this client. It is normmaly listed as the last Message type to
 * in the interested message type list to catch all
 * - see Client.hpp and ConsoleClient.hpp documentation
 */
struct JustBytes
: hasTag<4> {
	JustBytes() = delete;
};

struct Trace {
    Trace() 
    : tsIndex(1) {
        ts[0] = (time::SysTime::now());
    }
    uint8_t tsIndex;
    time::SysTime ts[6];
    friend 
    std::ostream& operator << (std::ostream& os, Trace const & t) {
        for (auto i = 0u; i < t.tsIndex; ++i) os << t.ts[i] << ' ';
        return os;
    }
};

template<>
struct MessageWrap<JustBytes> :  MessageHead {
	MessageWrap(uint16_t tag, void const* bytes, size_t len)
	: MessageHead(tag) {
		memcpy(&payload, bytes, len);
	}
    uint8_t payload;

    friend 
    std::ostream& operator << (std::ostream& os, MessageWrap<JustBytes> const & w) {
    	return os << static_cast<MessageHead const&>(w) << " *";
    }
};
static_assert(sizeof(MessageWrap<JustBytes>) == sizeof(MessageHead) + 1, "wrong packing");
static_assert(sizeof(MessageWrap<Flush>) == sizeof(MessageHead) + sizeof(Flush), "wrong packing");

/**
 * @class hasMemoryAttachment
 * @brief if a specific hmbdc network transport (for example tcpcast, rmcast, and rnetmap) supports message 
 * with memory attachment, the message needs to be derived from this base - as the FIRST base, 
 * so it can be handled properly by the hmbdc network transport when sending and receiving it
 * @details user on the sending side cannot directly free the attachment, instead the user can provide 
 * a callback func and hmbdc will call it - the default callback function is to call free on the attachment
 */
struct hasMemoryAttachment {
    using AfterConsumedCleanupFunc = void (*)(hasMemoryAttachment*);

    /**
     * @brief default ctor 
     * @details will allocate and release thru heap
     */
    hasMemoryAttachment()
    : afterConsumedCleanupFunc(hasMemoryAttachment::free)
    , attachment(NULL)
    , len(0) {
    }
    enum {
        flag = 1
    };

    /**
     * @brief file mapped memory
     * @param fileName file name to map
     */
    hasMemoryAttachment(char const* fileName) 
    : afterConsumedCleanupFunc(nullptr)
    , attachment(nullptr)
    , len(0) {
        map(fileName);
    }

    void release() {
        if (afterConsumedCleanupFunc) {
            afterConsumedCleanupFunc(this);
            afterConsumedCleanupFunc = nullptr;
        }
    }
   
    /**
     * @brief map this object's attached memory to a file
     * @details the afterConsumedCleanupFunc is automatically set to do the unmap
     * 
     * @param fileName file to map into memory
     * @return size of the file mapped
     */
    size_t map(char const* fileName);
    AfterConsumedCleanupFunc afterConsumedCleanupFunc; /// pointing to a func that handles cleaning up the attachment
                                                       /// it is automatically set, but user can change it as desire
                                                       /// when the sending side is done with this message, 
                                                       /// this is called by hmbdc automatically.
                                                       /// however, this is not called on the recv side since 
                                                       /// only the user code knows when to call it

    /**
     * @brief does the reverse of map
     */
    void 
    unmap() {
        unmap(this);
    }
    void* attachment;
    XmitEndian<size_t> len;

    friend 
    std::ostream& operator << (std::ostream& os, hasMemoryAttachment const & m) {
        return os << "hasMemoryAttachment " << m.attachment << ' ' << m.len;
    }

    static
    void 
    unmap(hasMemoryAttachment*);

    static
    void 
    free(hasMemoryAttachment* a) {
        ::free(a->attachment);
        a->attachment = nullptr;
    }
} __attribute__((aligned (8)));

struct StartMemorySegTrain
: hasTag<5> {
    XmitEndian<uint16_t> underlyingMessageTypeTag;
    hasMemoryAttachment att;
    XmitEndian<size_t> segCount;

    friend 
    std::ostream& operator << (std::ostream& os, StartMemorySegTrain const & m) {
        return os << "StartAttachmentTrain " << m.underlyingMessageTypeTag 
        << ' ' << m.att << ' ' << m.segCount;
    }
} __attribute__((aligned(8)));

struct MemorySeg
: hasTag<6> {
    void const *  seg;
    XmitEndian<uint16_t> len;
    friend 
    std::ostream& operator << (std::ostream& os, MemorySeg const & m) {
        return os << "MemorySeg " << m.seg << ' ' << m.len;
    }
} __attribute__((packed));

/**
 * @class LastSystemMessage
 * @brief hmbdc system messages use tag values less than this one
 */
struct LastSystemMessage 
: hasTag<999> {
};

/**
 * @brief      template that convert a regular message to be a request 
 * used in request / reply sync messaging - see RequestReply.hpp for details
 *
 * @tparam     Message  a regular message type 
 */
template <MessageC Message>
struct REQUEST
: hasTag<Message::typeTag> {
    comm::Uuid uuid;
    Message request;

    template <typename ...Args>
    explicit REQUEST(Args&&... args)
    : request(std::forward<Args>(args)...)
    {}

    REQUEST(REQUEST const&) = default;
    REQUEST(REQUEST &) = default;
    REQUEST(REQUEST &&) = default;
    REQUEST& operator = (REQUEST const&) = default;
    REQUEST& operator = (REQUEST &&) = default;
};

/**
 * @brief      template that convert a regular message to be a reply 
 * used in request / reply sync messaging - see RequestReply.hpp for details
 *
 * @tparam     Message  a regular message type 
 */
template <MessageC Message>
struct REPLY
: hasTag<Message::typeTag> {
    template <typename Request, typename ...Args>
    explicit REPLY(REQUEST<Request> const& req
        , Args&&... args)
    : uuid(req.uuid)
    , reply(std::forward<Args>(args)...)
    {}

    comm::Uuid uuid;
    Message reply;
};

template <>
struct REPLY<JustBytes>
: hasTag<7> {
    comm::Uuid uuid;
    char reply;
};

template<>
struct MessageWrap<REPLY<JustBytes>> :  MessageHead {
    MessageWrap(uint16_t tag, void const* bytes, size_t len)
    : MessageHead(tag) {
        memcpy(&payload, bytes, len);
    }
    REPLY<JustBytes> payload;

    friend 
    std::ostream& operator << (std::ostream& os, MessageWrap<REPLY<JustBytes>> const & w) {
        return os << static_cast<MessageHead const&>(w) << " *";
    }
};

#if __GNUC_PREREQ(9,0)

template <typename R>
concept RequestC = hmbdc::is_template<REQUEST, typename std::decay<R>::type>;
template <typename R>
concept ReplyC = hmbdc::is_template<REPLY, typename std::decay<R>::type>;

#else

#define RequestC typename
#define ReplyC typename

#endif

}} //hmbdc::app
