#include "hmbdc/Copyright.hpp"
#pragma once

#include "hmbdc/comm/Uuid.hpp"
#include "hmbdc/Exception.hpp"


#include <functional>
#include <mutex>
#include <tuple>
#include <memory>
#include <condition_variable>

namespace hmbdc { namespace app {

namespace request_reply_proxy_client_detail {
template <typename CcClient, typename Message>
struct dispatcher {
    bool operator()(CcClient& c, MessageHead& m, comm::Uuid const& uuid) {
        if (m.typeTag == Message::typeTag 
        	&& uuid == m.get<Message>().uuid) {
            c.handleReplyCb(m.get<Message>().reply);
            using Wrap = MessageWrap<Message>;
            auto& wrap = static_cast<Wrap&>(m);
            wrap.~Wrap();
            return true;
        }
        return false;
    }
};


template <typename CcClient>
struct dispatcher<CcClient, REPLY<JustBytes>> {
    bool operator()(CcClient& c, MessageHead& m, comm::Uuid const& uuid) {
        //only deliver user messages
        if (m.typeTag > 1000 && uuid == m.get<REPLY<JustBytes>>().uuid) { 
            auto j = &m.get<REPLY<JustBytes>>().reply;
            c.handleJustBytesReplyCb(m.typeTag, j);
        }
        return true;
    }
};

template <typename CcClient, typename ... Messages>
struct ReplyDispatcher {
    bool operator()(CcClient&, MessageHead&, comm::Uuid const&) const{return false;}
};

template <typename CcClient, typename M, typename ... Messages>
struct ReplyDispatcher<CcClient, std::tuple<M, Messages...>> {
    bool operator()(CcClient& c, MessageHead& e, comm::Uuid const& uuid) const {
        if (!dispatcher<CcClient, M>()(c, e, uuid))
            return ReplyDispatcher<CcClient, std::tuple<Messages ...>>()(c, e, uuid);
        else
            return true;
    }
};

struct SyncData {
    std::function<void()> send;
    bool proxyReady = false;
    bool allRepliesReceived = false;
    std::mutex mux;
    std::condition_variable cond;
};

template <typename CcClient>
struct RequestReplyProxyClient final {
    using Interests = typename CcClient::Replies;
    using Replies = std::tuple<>;
    using Requests = std::tuple<>;

    enum {
        INTERESTS_SIZE = std::tuple_size<Interests>::value,
    };
    char const* hmbdcName() const { return nullptr; }
    std::tuple<char const*, int> schedSpec() const {
        return std::make_tuple<char const*, int>(nullptr, 0);
    }
    size_t maxBatchMessageCount() const {return std::numeric_limits<size_t>::max();}
    void messageDispatchingStartedCb(uint16_t threadSerialNumber) {
        syncData->send();
    }
    void stoppedCb(std::exception const& e) {
    }
    bool droppedCb() {delete this; return true;};
    virtual ~RequestReplyProxyClient(){}

    RequestReplyProxyClient(CcClient& c)
    : syncData(std::make_shared<SyncData>())
    , uuid(comm::Uuid::generate())
    , c_(c) {
    }

    std::shared_ptr<SyncData> syncData;

    void invokedCb(uint16_t threadId) {
        if (syncData.use_count() == 1) {
            throw ExitCode(1);
        }
    }
    
    template <typename Iterator>
    void handleRangeImpl(Iterator it, 
        Iterator end, uint16_t threadId) {
        try {
	        for (;hmbdc_likely(it != end); ++it) {
	            request_reply_proxy_client_detail::ReplyDispatcher<CcClient, Interests>()(
	                c_, *static_cast<MessageHead*>(*it), uuid);
	        }
    	} catch (...) {
            std::unique_lock<std::mutex> lk(syncData->mux);
            syncData->allRepliesReceived = true;
    		syncData->cond.notify_one();
    		throw;
    	}
    }

    const comm::Uuid uuid;
private:
	CcClient& c_;
};

struct deliverBasedOnUuid {
	comm::Uuid const* uuid;
    template <typename ReplyT>
    bool operator()(ReplyT&& t) {return t.uuid == *uuid;}

};

}

/**
 * @brief      not an interface exposed to end user
 */
template <RequestC RequestT, typename Sender, typename ReplyContext, typename ReplyHandler
>
bool
requestReply(RequestT &&request
	, Sender& sender
	, ReplyContext& ctx
    , ReplyHandler& replyHandler
    , time::Duration timeout) {
    static_assert(ReplyContext::cpa::has_pool && !ReplyContext::cpa::pool_msgless
        , "Contex needs to have a pool");
	using namespace request_reply_proxy_client_detail;
    auto proxy = new RequestReplyProxyClient<ReplyHandler>(replyHandler);
    auto syncData = proxy->syncData;
    request.uuid = proxy->uuid;
    syncData->send = [&sender, &request]() {
        sender.send(std::forward<RequestT>(request));
    };
    ctx.addToPool(*proxy);
    std::unique_lock<std::mutex> lk(syncData->mux);
    return syncData->cond.wait_for(lk, std::chrono::microseconds(timeout.microseconds())
        , [syncData]() {return syncData->allRepliesReceived;});
}

/**
 * @brief synchronous request reply interface for the Network Sender
 * 
 * @tparam CcSender concrete Sender
 */
template <typename CcSender>
struct RequestReplySender {
    /**
     * @brief The caller is blocked until the replies to this request are heard and 
     * processed. Just like regular Messages REQUEST is sent out to Clients through 
     * the CcSender
     * 
     * @details 
     * 
     * @tparam MessageC Message the type wrapped in REQUEST<> template
     * @tparam ReplyContext the Context Type that ReplyT messages need to be able 
     * to reach, it needs to have a pool
     * @snippet blind-auction.cpp nework request reply example
     * @tparam ReplyHandler type for how to handle the replies.
     *      example of a handler expect two steps replies for requestReply call:
     *      @code
     *      struct ClientAlsoHandlesReplies
     *      : Client<ClientAlsoHandlesReplies, REPLY<Reply0>, REPLY<Reply1>, OtherMsg> {
     *          void handleReplyCb(Reply0 const&){
     *              // got step 1 reply
     *          }
     *          void handleReplyCb(Reply1 const&){
     *              // got step 2 reply 
     *              throw hmbdc::ExitCode(0); // all replies received, now
     *                                        // throw to unblock the requestReply call
     *          }
     *          @endcode
     * 
     * @param request the REQUEST sent out
     * @param ctx the replies are expected to show up in this Context which contains a pool
     * @param replyHandler the callback logic to handle replies
     */
    template <RequestC RequestT, typename ReplyContext, typename ReplyHandler>
    bool
    requestReply(RequestT&& request
        , ReplyContext& ctx
        , ReplyHandler& replyHandler
        , time::Duration timeout = time::Duration::seconds(0xffffffff)) {
        static_assert(ReplyContext::cpa::has_pool && !ReplyContext::cpa::pool_msgless
            , "not supported - needs to have a pool");
        return hmbdc::app::requestReply(std::forward<RequestT>(request)
            , *static_cast<CcSender*>(this)
            , ctx
            , replyHandler
            , timeout);
    }


    /**
     * @brief part of synchronous request reply interface. Send a constructed REPLY
     * using the CcSender, it will only reach to the origianl sender Client. 
     * ALL REPLY via a Sender need to sent out using this interface to avoid
     * undefined behavior.
     * 
     * @tparam MessageC Message the type wrapped in REPLY<> template
     *      example usage:
     *      @code
     *      REPLY<Response> resp(req, req.request); //see REPLY<> ctor
     *      myCtx.sendReply(resp);
     *      @endcode
     * 
     * @param reply (rvalue) the REPLY (previously constructed from a REQUEST) to sent out
     */
    template <MessageC M>
    void
    sendReply(REPLY<M> && reply) {
        (static_cast<CcSender*>(this))->send(std::forward<REPLY<M>>(reply));
    }

    /**
     * @brief part of synchronous request reply interface. Send a constructed REPLY
     * using the CcSender, it will only reach to the origianl sender Client. 
     * ALL REPLY via a Sender need to sent out using this interface to avoid
     * undefined behavior.
     * 
     * @tparam MessageC Message the type wrapped in REPLY<> template
     *      example usage:
     *      @code
     *      REPLY<Response> resp(req, req.request); //see REPLY<> ctor
     *      myCtx.sendReply(resp);
     *      @endcode
     * 
     * @param reply (lvalue) the REPLY (previously constructed from a REQUEST) to sent out
     */
    template <MessageC M>
    void
    sendReply(REPLY<M> const& reply) {
        static_cast<CcSender*>(this)->send(reply);
    }
};

}}


