#include "hmbdc/Copyright.hpp"
#pragma once
#include "hmbdc/app/Message.hpp"
#include "hmbdc/pattern/GuardedSingleton.hpp"

namespace hmbdc { namespace app { namespace utils {

namespace netportal_detail {
template <typename MesssageRecver, MessageC... RecvMessages>
struct Proxy : Client<Proxy<MesssageRecver, RecvMessages...>, RecvMessages ...> {
    Proxy(MesssageRecver&& recver)
    : recver(std::forward<MesssageRecver>(recver)){}
    MesssageRecver&& recver;

    template <MessageC Message>
    void handleMessageCb(Message&& message) {
        count++;
        recver.handleMessageCb(std::forward<Message>(message));
    }
    size_t count = 0;
};
} //detail

/**
 * @brief an utility that gives finer control to the user that do
 * not rely on hmbdc Contexts
 * @details still needs to use NetContext singleton which needs 
 * to be created first
 * 
 * @tparam NetContext - on the hmbdc net transport Context such as
 * tcpcast::NetContext, ... etc
 * @tparam SendMessageTuple - std tuple containing the message types to 
 * send out - can be empty
 * @tparam RecvMessageTuple - std tuple containing the message types to 
 * receive - default to be empty
 * 
 */
template <typename NetContext, typename SendMessageTuple
    , typename RecvMessageTuple = std::tuple<>>
struct NetPortal;

template <typename NetContext, MessageC ... SendMessages, MessageC ... RecvMessages>
struct NetPortal<NetContext, std::tuple<SendMessages...>, std::tuple<RecvMessages...>> { 
    /**
     * @brief ctor - only can be called after the NetContext singleton is created
     * @details it does not go thru the state machine for sender / receiver
     * hand shake if the netowork transport dictates that
     * calling rotate several times will make it happen
     * 
     * @param cfg a json configure that has a tx and / or rx sections
     * the supported params are the same as individual transport
     * @param messageQueueSizePower2Num - see doc in NetContext
     */    
    NetPortal(Config cfg, size_t messageQueueSizePower2Num = 20)
    : ctx_(messageQueueSizePower2Num, 0)
    , sendEng_(sizeof...(SendMessages)?NetContext::instance().createSendTransportEngine(
        (cfg.put("connKeyCheck", false), cfg.resetSection("tx", false))
        , max_size_in_tuple<std::tuple<SendMessages...>>::value)
        : nullptr)
    , recvEng_(sizeof...(RecvMessages)?NetContext::instance().createRecvTransportEngine(
        cfg.resetSection("rx", false)
        , ctx_.buffer())
        : nullptr)
    {}

    /**
     * @brief run the portal to send and receive messages
     * @details it does NOT block - it is not garanteed the user
     * will have sent or received messages. It is expected to be called
     * multiple times
     * thread safe
     */
    void rotate() {
        if (hmbdc_likely(recvEng_)) recvEng_->rotate();
        if (hmbdc_likely(sendEng_)) sendEng_->rotate();
    }

    /**
     * @brief just run the portal part that receives messages
     * @details it does NOT block - it is not garanteed the user
     * will have received messages. It is expected to be called
     * multiple times
     * thread safe
     * 
     * @tparam MesssageRecver a type that has the handleMessageCb()
     * callbacks - see app::Client's doc for those callbacks
     * @param recver an instance of the receiver
     * @return the message count that has been dispatched to the recver
     * in this call
     */
    template <typename MesssageRecver>
    size_t pullIn(MesssageRecver&& recver) {
        using Proxy = netportal_detail::Proxy<MesssageRecver, RecvMessages...>;
        if (hmbdc_likely(recvEng_)) recvEng_->rotate();

        Proxy proxy(std::forward<MesssageRecver>(recver));
        ctx_.runClientThreadOnce(0, proxy);
        return proxy.count;
    }

    /**
     * @brief just run the portal part that send messages
     * @details it does NOT block - it is not garanteed the user
     * will have sent messages depending on underly network transport. 
     * It is expected to be called multiple times
     * thread safe
     * 
     * @details in low traffic cases the messages are always sent out
     * by single call
     */
    void flushOut() {
        if (hmbdc_likely(sendEng_)) sendEng_->rotate();
    }

    /**
     * @brief get a sender for a topic
     * @details thread safe but a slow call 
     * - so the user cashing the result is expected
     * 
     * @param t topic for sending
     * @return sender point. The sender is owned by the lib - do not delete
     */
    typename NetContext::Sender* getSender(comm::Topic const& t) {
        if (sendEng_ && sendEng_->match(t))
            return NetContext::instance().getSender(t);
        return nullptr;
    }

    /**
     * @brief listen to a topic - has impact on the
     * pullIn call
     * @details threadsafe
     * 
     * @param t topic to listen to
     */
    void listenTo(comm::Topic const& t) {
        if (recvEng_) recvEng_->listenTo(t);
    }

    /**
     * @brief stop listening to a topic - has impact on the
     * pullIn call
     * @details threadsafe
     * 
     * @param t topic to forget
     */
    void stopListenTo(comm::Topic const& t) {
        if (recvEng_) recvEng_->stopListenTo(t);
    }

    /**
     * @brief the sending session that are active
     * which is how many receiver are active
     * @details only supported for reliable transport types
     * such as tcpcast, rmast, and rnemap
     * @return the count - return numeric_limits<size_t>::max()
     * when the first receiver is yet to connect (initial)
     */
    size_t sendSessionsRemainingActive() const {
        return sendEng_?sendEng_->sessionsRemainingActive():0;
    }
private:
    using Ctx 
        = Context<max_size_in_tuple<std::tuple<RecvMessages...>>::value
            , context_property::partition
        >;
    Ctx ctx_;
    using SendTransportEngine = typename NetContext::SendTransportEngine;
    SendTransportEngine* sendEng_;
    using RecvTransportEngine = typename NetContext::template RecvTransportEngine<typename Ctx::Buffer>;
    RecvTransportEngine* recvEng_;
};

}}}

